/* life.c */

#include <Pilot.h>
#include "callback.h"

#include "life.h"
#include "life_rcp.h"


/*
 * types and macros
 */

#define MAX_LENGTH(max) (((max) - ((max) / 6 + 4096)) / sizeof(cell_t))

#define SCROLL_MARGIN 24
#define SCROLL_OFFSET 16


/*
 * global variables
 */

static cell_t *tape = NULL;
static cell_t *tape_begin;
static cell_t *tape_end;
static offset_t tape_length;
static offset_t init_length;
static const cell_t *life_begin;
static const cell_t *life_end;

static gen_t generation;
static gen_t until;
static Boolean time;
static gen_t from_gen;
static ULong from_ticks;

static coord_t scroll_x;
static coord_t scroll_y;
static Boolean running;
static Boolean can_back;

static char gens[6];
static char cells[6];

static DmOpenRef memo_db;
static UInt life_category;
static ListPtr list;
static Word last_selection;
static UInt last_top;


/*
 * common routines
 */

static VoidPtr GetObjectPtr(Word id)
{
    FormPtr form;

    form = FrmGetActiveForm();
    return FrmGetObjectPtr(form, FrmGetObjectIndex(form, id));
}

static void SetFieldNumber(FieldPtr field, CharPtr str, long number)
{
    StrIToA(str, number);
    FldSetTextPtr(field, str);
    FldDrawField(field);
}

static Word FindCategory(DmOpenRef db_ref, CharPtr category_name)
{
    char name[dmCategoryLength];
    Word i;

    for (i = 0; i < dmRecNumCategories; ++i) {
	CategoryGetName(db_ref, i, name);
	if (StrCaselessCompare(name, category_name) == 0) {
	    break;
	}
    }
    return i;
}


/*
 * init, free and reset life cells
 */

static void init_life()
{
    ULong free, max;

    MemHeapFreeBytes(0, &free, &max);
    tape_length = MAX_LENGTH(max) * 2;

    while (tape == NULL) {
	tape_length /= 2;
	tape = MemPtrNew(sizeof(cell_t) * tape_length);
    }

    tape_begin = tape;
    tape_end = tape + tape_length;

    init_length = 0;
    *tape = 0;
    life_begin = life_end = tape + 1;
    generation = 0;

    running = false;
}

static void free_life()
{
    if (tape != NULL) {
	MemPtrFree(tape);
    }
}

static void reset_life()
{
    tape[init_length] = NULL;
    tape_begin = tape + init_length;
    life_begin = tape_begin + 1;
    life_end = life_begin + init_length;
    MemMove((cell_t *) life_begin, tape, sizeof(cell_t) * init_length);
    generation = 0;

    scroll_x = gadget_x + gadget_width / 2;
    scroll_y = gadget_y + gadget_height / 2;
    running = false;
}

offset_t count_length(const cell_t *begin, const cell_t *end)
{
    if (begin <= end) {
	return end - begin;
    } else {
	return (end - tape_begin) + (tape_end - begin);
    }
}

const cell_t *incr_ptr(const cell_t *ptr)
{
    return (ptr + 1 == tape_end) ? tape_begin : ptr + 1;
}

const cell_t *decr_ptr(const cell_t *ptr)
{
    return (ptr == tape_begin) ? tape_end - 1 : ptr - 1;
}


/*
 * loading life data from memo
 */

static void draw_list_item(Word item_no, RectanglePtr bounds, CharPtr *dummy)
{
    UInt index = 0;

    CALLBACK_PROLOGUE

    if (DmSeekRecordInCategory(
	    memo_db, &index, item_no, dmSeekForward, life_category) == 0) {
	VoidHand handle = DmQueryRecord(memo_db, index);
	Int width = bounds->extent.x;
	Int length = MemHandleSize(handle);
	CharPtr ptr = MemHandleLock(handle);
	Boolean fits;

	FntCharsInWidth(ptr, &width, &length, &fits);
	WinDrawChars(ptr, length, bounds->topLeft.x, bounds->topLeft.y);

	MemHandleUnlock(handle);
    }

    CALLBACK_EPILOGUE
}

static void scroll_list(DirectionType direction)
{
    int top;

    if (direction == up) {
	top = list->topItem - (list_visible_items - 1);
	if (top < 0) {
	    top = 0;
	}
    } else {
	top = list->topItem + (list_visible_items - 1);
	if (top > LstGetNumberOfItems(list) - list_visible_items) {
	    top = LstGetNumberOfItems(list) - list_visible_items;
	}
    }

    if (top != list->topItem) {
	LstSetTopItem(list, top);
	LstEraseList(list);
	LstDrawList(list);
    }
}

static Boolean handle_load_dialog_event(EventPtr event)
{
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (event->eType) {
    case keyDownEvent:
	if (event->data.keyDown.chr == pageUpChr) {
	    scroll_list(up);
	    handled = true;
	} else if (event->data.keyDown.chr == pageDownChr) {
	    scroll_list(down);
	    handled = true;
	}
	break;

    default:
	break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static void load_life()
{
    UInt numRecords;
    FormPtr dialog;
    Word button, selection;

    memo_db = DmOpenDatabaseByTypeCreator('DATA', 'memo', dmModeReadOnly);
    if (memo_db == 0) {
	FrmAlert(load_error1_id);
	return;
    }

    life_category = FindCategory(memo_db, "Life");
    if (life_category >= dmRecNumCategories) {
	DmCloseDatabase(memo_db);
	FrmAlert(load_error2_id);
	return;
    }

    numRecords = DmNumRecordsInCategory(memo_db, life_category);
    if (numRecords == 0) {
	DmCloseDatabase(memo_db);
	FrmAlert(load_error3_id);
	return;
    }

    dialog = FrmInitForm(load_form_id);
    list = FrmGetObjectPtr(dialog, FrmGetObjectIndex(dialog, load_list_id));
    LstSetDrawFunction(list, &draw_list_item);
    LstSetListChoices(list, NULL, numRecords);
    LstSetSelection(list, last_selection);
    LstSetTopItem(list, last_top);
    FrmSetEventHandler(dialog, handle_load_dialog_event);
    button = FrmDoDialog(dialog);
    selection = LstGetSelection(list);
    last_selection = selection;
    last_top = list->topItem;
    FrmDeleteForm(dialog);

    if (button == load_ok_id && selection != noListSelection) {
	UInt index = 0;

	if (DmSeekRecordInCategory(memo_db,
		&index, selection, dmSeekForward, life_category) == 0) {
	    VoidHand handle = DmQueryRecord(memo_db, index);
	    CharPtr ptr = MemHandleLock(handle);
	    offset_t length;

	    length = load(ptr, MemHandleSize(handle), NULL);

	    if (length <= 0) {
		FrmAlert(load_error4_id);
	    } else if (length >= tape_length / 2 - 1) {
		FrmAlert(load_error5_id);
	    } else {
		init_length = load(ptr, MemHandleSize(handle), tape);
		reset_life();
	    }

	    MemHandleUnlock(handle);
	}
    }

    DmCloseDatabase(memo_db);
}


/*
 * drawing, calculating and scrolling life cells
 */

static void draw_cells(Boolean update)
{
    const cell_t *ptr, *prev_begin, *prev_end, *prev_ptr;
    cell_t saved_end, cell, prev_cell, sentinel;
    coord_t x, y;

    if (! update) {
	RectangleType r;

	RctSetRectangle(&r,
	    gadget_x + 1, gadget_y + 1, gadget_width - 2, gadget_height - 2);
	WinDrawRectangleFrame(rectangleFrame, &r);
	WinEraseRectangle(&r, 0);

	prev_begin = prev_end = &sentinel;
    } else {
	prev_end = decr_ptr(life_begin);
	prev_begin = (const cell_t *) *prev_end;
    }

    saved_end = *life_end;
    *(cell_t *) life_end = SENTINEL;
    *(cell_t *) prev_end = SENTINEL;

    cell = *(ptr = life_begin);
    prev_cell = *(prev_ptr = prev_begin);

    while (true) {
	if (cell < prev_cell) {
	    x = GET_X(cell) + scroll_x;
	    y = GET_Y(cell) + scroll_y;
	    cell = *(ptr = incr_ptr(ptr));
	    goto invert_cell;
	} else if (cell > prev_cell) {
	    x = GET_X(prev_cell) + scroll_x;
	    y = GET_Y(prev_cell) + scroll_y;
	    prev_cell = *(prev_ptr = incr_ptr(prev_ptr));
	invert_cell:
	    if (x >= gadget_x + 1 && x < gadget_x + gadget_width  - 1
	     && y >= gadget_y + 1 && y < gadget_y + gadget_height - 1) {
		WinInvertLine(x, y, x, y);
	    }
	} else if (cell != SENTINEL) {
	    cell = *(ptr = incr_ptr(ptr));
	    prev_cell = *(prev_ptr = incr_ptr(prev_ptr));
	} else {
	    break;
	}
    }

    *(cell_t *) life_end = saved_end;
    *(cell_t *) prev_end = (cell_t) prev_begin;

    SetFieldNumber((FieldPtr) GetObjectPtr(main_gens_id), gens, generation);
    SetFieldNumber((FieldPtr) GetObjectPtr(main_cells_id),
	cells, count_length(life_begin, life_end));
}

static void draw_go_stop_button()
{
    VoidHand h;
    VoidPtr p;

    h = DmGetResource('Tbmp', running ? stop_bitmap_id : go_bitmap_id);
    if (h != NULL) {
	p = MemHandleLock(h);
	WinDrawBitmap(p, main_go_x, main_go_y);
	MemHandleUnlock(h);
	DmReleaseResource(h);
    }
}

static void draw_back_button(Boolean forced)
{
    Boolean new_back;
    VoidHand h;
    VoidPtr p;

    new_back = (*decr_ptr(life_begin) != NULL);
    if (forced || new_back != can_back) {
	can_back = new_back;
	CtlSetUsable(GetObjectPtr(main_back_id), can_back);
	CtlSetUsable(GetObjectPtr(main_fback_id), can_back);

	h = DmGetResource('Tbmp',
		can_back ? back_bitmap_id : cant_back_bitmap_id);
	if (h != NULL) {
	    p = MemHandleLock(h);
	    WinDrawBitmap(p, main_back_x, main_back_y);
	    MemHandleUnlock(h);
	    DmReleaseResource(h);
	}
    }
}

static void set_running(Boolean run)
{
    if (run != running) {
	ControlPtr go_btn = (ControlPtr) GetObjectPtr(main_go_id);
	ControlPtr stop_btn = (ControlPtr) GetObjectPtr(main_stop_id);

	CtlEraseControl(run ? go_btn : stop_btn);
	CtlSetUsable(go_btn, ! run);
	CtlSetUsable(stop_btn, run);
	CtlDrawControl(run ? stop_btn : go_btn);
	running = run;

	draw_go_stop_button();
    }
}

static void go_until()
{
    FormPtr dialog;
    FieldPtr field;
    Word index, button;
    VoidHand h;
    CharPtr ptr;

    dialog = FrmInitForm(until_form_id);
    index = FrmGetObjectIndex(dialog, until_field_id);
    FrmSetFocus(dialog, index);
    field = FrmGetObjectPtr(dialog, index);
    h = MemHandleNew(12);
    if (h != NULL) {
	ptr = MemHandleLock(h);
	StrIToA(ptr, generation);
	MemHandleUnlock(h);
	FldFreeMemory(field);
	FldSetTextHandle(field, (Handle) h);
	FldSetSelection(field, 0, FldGetTextLength(field));
    }

    button = FrmDoDialog(dialog);
    if (button == until_ok_id
     && (ptr = FldGetTextPtr(field)) != NULL && *ptr != '\0') {
	until = StrAToI(ptr);
	if (until < generation) {
	    until = generation;
	}
	index = FrmGetObjectIndex(dialog, until_time_id);
	time = CtlGetValue(FrmGetObjectPtr(dialog, index));
	from_gen = generation;
	from_ticks = TimGetTicks();
    }
    FrmDeleteForm(dialog);

    if (until > generation) {
	set_running(true);
    } else if (until == generation) {
	set_running(false);
    }
}

static void show_time()
{
    char from[6], to[6], ticks[12];

    StrIToA(from, from_gen);
    StrIToA(to, generation);
    StrIToA(ticks, TimGetTicks() - from_ticks);
    FrmCustomAlert(time_alert_id, from, to, ticks);
}

static void next_gen(Boolean need_check)
{
    const cell_t *begin, *end;
    cell_t *ptr;

    if (life_end == life_begin) {
	set_running(false);
	draw_back_button(false);
	return;
    }

    begin = incr_ptr(life_end);
    ptr = (cell_t *) decr_ptr(life_begin);
    end = merge(tape_begin, tape_end, life_begin, life_end,
		(cell_t *) begin, ptr);
    if (end == NULL) {
	*ptr = NULL;
	set_running(false);
	draw_back_button(false);
	FrmAlert(outof_error_id);
	return;
    }

    *(cell_t *) life_end = (cell_t) life_begin;

    life_begin = begin;
    life_end = end;

    for ( ; *ptr != NULL; ptr = (cell_t *) begin) {
	begin = decr_ptr((const cell_t *) *ptr);
	if ((begin < ptr && (end > begin && end <= ptr))
	 || (begin > ptr && (end <= ptr || end > begin))) {
	    break;
	}
    }
    *ptr = NULL;

    if (need_check) {
	int count, num_gliders;
	char gens[6], from[6], gliders[6];

	count = check(life_begin, life_end, &num_gliders);
	if (count > 0) {
	    life_end = decr_ptr(life_begin);
	    life_begin = (const cell_t *) *life_end;
	    set_running(false);

	    StrIToA(gens, count);
	    StrIToA(from, generation + 1 - count);
	    if (num_gliders == 0) {
		FrmCustomAlert(loop_alert_id, gens, from, " ");
	    } else {
		StrIToA(gliders, num_gliders);
		FrmCustomAlert(loop2_alert_id, gens, from, gliders);
	    }

	    return;
	}
    }

    ++generation;

    draw_back_button(false);
    draw_cells(true);
}

static void go_back()
{
    const cell_t *ptr = decr_ptr(life_begin);

    if (*ptr != NULL) {
	--generation;

	draw_cells(true);

	life_begin = (const cell_t *) *ptr;
	life_end = ptr;

	draw_back_button(false);
    }
}

static Boolean scroll(EventPtr event)
{
    coord_t x = event->screenX;
    coord_t y = event->screenY;

    if (x >= gadget_x && x < gadget_x + gadget_width
     && y >= gadget_y && y < gadget_y + gadget_height) {
	if (x < gadget_x + SCROLL_MARGIN) {
	    scroll_x += SCROLL_OFFSET;
	} else if (x >= gadget_x + gadget_width - SCROLL_MARGIN) {
	    scroll_x -= SCROLL_OFFSET;
	}
	if (y < gadget_y + SCROLL_MARGIN) {
	    scroll_y += SCROLL_OFFSET;
	} else if (y >= gadget_y + gadget_height - SCROLL_MARGIN) {
	    scroll_y -= SCROLL_OFFSET;
	}
	return true;
    }
    return false;
}


/*
 * event handlers
 */

static Boolean handle_main_form_event(EventPtr event)
{
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (event->eType) {
    case frmOpenEvent:
	FrmDrawForm(FrmGetActiveForm());
	reset_life();
	handled = true;

	/* fall through */

    case frmUpdateEvent:
	draw_go_stop_button();
	draw_back_button(true);
	draw_cells(false);
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case main_go_id:
	    until = -1;
	    set_running(true);
	    handled = true;
	    break;

	case main_stop_id:
	    set_running(false);
	    handled = true;
	    break;

	case main_step_id:
	    set_running(false);
	    next_gen(false);
	    handled = true;
	    break;

	case main_back_id:
	    set_running(false);
	    go_back();
	    handled = true;
	    break;

	case main_until_id:
	    go_until();
	    handled = true;
	    break;

	case main_reset_id:
	    set_running(false);
	    reset_life();
	    draw_back_button(false);
	    draw_cells(false);
	    handled = true;
	    break;

	case main_load_id:
	    set_running(false);
	    load_life();
	    draw_back_button(false);
	    draw_cells(false);
	    handled = true;
	    break;

	case main_about_id:
	    FrmAlert(about_alert_id);
	    handled = true;
	    break;
	}
	break;

    case ctlRepeatEvent:
	switch (event->data.ctlRepeat.controlID) {
	case main_fstep_id:
	    set_running(false);
	    next_gen(false);
	    break;

	case main_fback_id:
	    set_running(false);
	    go_back();
	    break;
	}
	break;

    case penDownEvent:
	if (scroll(event)) {
	    draw_cells(false);
	    handled = true;
	    break;
	}
	break;

    case nilEvent:
	if (running) {
	    next_gen(true);
	    if (until >= 0 && generation >= until) {
		set_running(false);
		if (time) {
		    show_time();
		}
	    }
	}
	handled = true;
	break;

    default:
	break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean handle_app_event(EventPtr event)
{
    FormPtr form;

    if (event->eType == frmLoadEvent) {
	form = FrmInitForm(event->data.frmLoad.formID);
	FrmSetActiveForm(form);
	if (event->data.frmLoad.formID == main_form_id) {
	    FrmSetEventHandler(form, handle_main_form_event);
	}
	return true;
    }
    return false;
}

static void event_loop()
{
    EventType event;
    Word error;

    do {
	EvtGetEvent(&event, running ? 1 : evtWaitForever);
	if (SysHandleEvent(&event) == false
	 && MenuHandleEvent(NULL, &event, &error) == false
	 && handle_app_event(&event) == false) {
	    FrmDispatchEvent(&event);
	}
    } while (event.eType != appStopEvent);
}


/*
 * main entry point
 */

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    if (cmd == sysAppLaunchCmdNormalLaunch) {
	init_life();
	FrmGotoForm(main_form_id);
	event_loop();
	FrmSaveAllForms();
	FrmCloseAllForms();
	free_life();
    }
    return 0;
}
